#!/usr/bin/env newlisp

"
Tags are evaluated by passing the data context as the first argument, any child
nodes (or an empty list) as the second argument, and a list of the parsed tag
arguments as the third argument (not including the tag name).

So, if you register a tag called to-upper that accepts a single string and makes
it upper-case and you use it with {% upper-case foo %}, it would be rendered by
evaluating:
  (TAGS:to-upper data-context child-nodes '(Variable foo))

Tags that take multiple arguments receive a final list with multiple arguments.
Arguments are lexed and are instances of Variable, Tag, String, or Comment
(tokens.lsp).  Each will have a :render method (for more control), or you can
render the entire list after modifying the data context using:
  (Template:render child-nodes ctx).

To register a new tag, define the tag-name as a function in the TAGS context:
  (define (TAGS:my-tag ctx children arg-list)
    ;; do something with ctx, perhaps adding values from arg-list or something
    (Template:render children))

Tags must evaluate to strings.
"

;;; Tag Library

(define (TAGS:for ctx rst)
  "for iterates over the list provided, redefining `element` on each iteration.
   The variable `counter` is also defined and is the zero-indexed count of
   iterations.  E.g.,
   
    {% for element in list %}
      {{ counter }}: {{ element }}
    {% endfor %}
  "
  (let ((elt (args 0 0)) (lst (args 0 2)) (buff ""))
    (set 'lst (if (context? ctx lst) (context ctx lst) '()))
    (letex (e (sym elt))
      (dolist (e lst)
        (context ctx elt e)
        (context ctx "counter" $idx)
        (write-buffer buff (Template:render rst ctx)))
      buff)))


(define (tag? str)
  (and (context? TAGS str)
       (lambda? (context TAGS str))))

