#!/usr/bin/env newlisp


;;; Author: Jeff Ober (http://www.artfulcode.net)
;;; Maintainer: Jeff Ober (http://www.artfulcode.net)
;;; Revision: 0.1 beta
;;; Released: 2008-02-26
;;;
;;; This software may be used by anyone for any purpose, so long as this header
;;; remains intact.


[text]

Template Module

The template module was inspired largely by the Django template system. I firmly
believe that separating complex logic from the template itself leads to both
cleaner code and, much more importantly, a simpler system for non-programmers. A
content producer should not be required to know lisp to write a simple template.

If you disagree, you are welcome to write your own template module ;)

Templates are rendered using data in a newLISP context. This is to preserve the
encapsulation of the view data and to prevent tags from accidentally accessing
external data. A custom tag or filter must explicitly access data from another
context in order to display it in the rendered template.

The primary module function is `Template:render`, which is called with a string
(the template text) and a context (the hash containing the view data). This is
tokenized and lexed by the module and rendered to a new string.

Tags may be defined as functions in the `TAGS` context. A tag is a function
which accepts first a view context, then a list of child nodes (parsed tokens
from the template text), then a list of the arguments passed to the tag (in
string format). The parsed tokens will each have a `render` method, but more
commonly `Template:render` will be used on the rest of the list.

Complex tags, such as `for` (below), iterate over a portion of the template (the
child nodes passed to the function), modifying the context on each iteration and
adding the rendered result to the string. Tags must be "ended", so a `for` tag
must end in and `endfor` tag.

	(define (TAGS:for ctx rst)
		;; (args) contains the words inside the tag tokens:
		;;   if the tag was called {% for i in x %}, then
		;;   (args) ; => '("i" "in" "x")
	  (let ((elt (args 0 0)) (lst (args 0 2)) (buff ""))
			;; get the data to iterate over (or use an empty list)
	    (set 'lst (if (context? ctx lst) (context ctx lst) '()))
	    (letex (e (sym elt))
	      (dolist (e lst)
					;; update the context with the current list item
	        (context ctx elt e)
					;; add the current loop iteration
	        (context ctx "counter" $idx)
					;; render the child-nodes using the modified context
	        (write-buffer buff (Template:render rst ctx)))
				;; we are left with the fully rendered string
	      buff)))

Simple tags will be passed an empty list in place of child nodes.

Currently, there is no checking to verify that a tag is used properly, nor is
there any safety built in to prevent incorrectly used tags being rendered. At
the moment, that is the responsibility of the developer.

Tags are used in the format:

 {% some-tag args passed to tag %}
 text to be parsed and passed to the tag
 function as child nodes.
 {% endsome-tag %}

Variables are tokens that appear between double curly braces, such as `{{
some-var }}`. If they appear in the view context, they will be rendered as their
string value. Otherwise, they will evaluate to an empty string (so that nil is a
valid value for a variable).

Variables may be further modified in-place using filters. Filter functions are
defined in the context `FILTERS` and accept up to two arguments, the string
value of the variable and a second string argument to the filter, in the form:

 {{ variable:filter:"some argument" }}

Filters must evaluate to a string.

Comments in a template appear between curly braces and hash marks:

 {# this is a comment.
    it may be multi-line. #}

White space is preserved in the template, except for that which appears inside
of comments. See tags.lsp, filters.lsp, and tokens.lsp for more detail on the
implementation of each.

[/text]


(load "tags.lsp")
(load "filters.lsp")
(load "tokens.lsp")


;;; Template is defined with several static functions and values.  Only render
;;; is truly a public function.


(context Template)

;;; The basic tokens can be configured
(set 'master-open-token "{")
(set 'master-close-token "}")


;;; These tokens are built from the basic tokens;
;;; The opening and closing tokens are defined here.
(set 'open-var (format "%s{" master-open-token))
(set 'close-var (format "}%s" master-close-token))
(set 'open-tag (format "%s%%" master-open-token))
(set 'close-tag (format "%%%s" master-close-token))
(set 'open-comment (format "%s#" master-open-token))
(set 'close-comment (format "#%s" master-close-token))


;;; These are conveniences to keep later code short; they
;;; combine the opening and closing tokens so that multiple
;;; tokens can be matched in one expression.
(set 'open-token-tokens (list open-var open-tag open-comment))
(set 'open-token (join open-token-tokens "|"))
(set 'close-token-tokens (list close-var close-tag close-comment))
(set 'close-token (join close-token-tokens "|"))


(define (extract str-token , pattern)
  "extract takes a string token and returns the word contents
  of the token."
  (let ((pattern
        (format "\%s(.)(?: )?(.*?)(?: )?\\1\%s"
                master-open-token
                master-close-token)))
       (regex pattern str-token) $2))


(define (tokenize str)
  "tokenize breaks string str into a list of string tokens."
  (cond
    ((empty? str) '())
    ;; parse variable
    ((= open-var (0 2 str))
     (letn ((m (find close-var str)) (next (+ m 2)))
           (cons (0 next str) (tokenize (next str)))))
    ;; parse tag
    ((= open-tag (0 2 str))
     (letn ((m (find close-tag str)) (next (+ 2 m)))
           (cons (0 next str) (tokenize (next str)))))
    ;; parse comment
    ((= open-comment (0 2 str))
     (letn ((m (find close-comment str)) (next (+ 2 m)))
           (cons (0 next str) (tokenize (next str)))))
    ;; parse text
    (true
     (letn ((m (find open-token str 0)))
           (if m
               (cons (0 m str) (tokenize (m str)))
               (cons str '()))))))


(define (lex tokens)
  "lex takes a list of tokens and creates a nested list of token objects."
  (cond
    ((null? tokens) '())
    ((= open-var (0 2 (first tokens)))
      (cons (Variable (extract (first tokens))) (lex (rest tokens))))
    ((= open-tag (0 2 (first tokens)))
      (letn ((opener (first (parse (extract (first tokens)))))
             (closer (format "%s( )?end%s( )?%s" open-tag opener close-tag))
             (found (find closer tokens 0)))
          (if found
            ;; collect entire phrase when there is a closing tag
            (cons (Tag (extract (first tokens)) (lex (1 (- found 1) tokens)))
                  (lex ((+ 1 found) tokens)))
            ;; otherwise it's a simple tag
            (cons (Tag (extract (first tokens)) (lex (rest tokens)))))))
    ((= open-comment (0 2 (first tokens)))
      (cons (Comment (extract (first tokens))) (lex (rest tokens))))
    (true
      ;; cons string to rest
      (cons (String (first tokens)) (lex (rest tokens))))))


(define (render hier ctx)
  "render has two syntaxes (or is it syntacti?):
    1. (render template-string data-context) => this renders the template using
       data from data-context.
    2. (render lexed-data data-context) => this is used from template tags to
       render any child-nodes with a modified context."
  (if (string? hier)
      (letn ((tokens (tokenize hier)) (lexed (lex tokens)))
            (set 'hier lexed)))
  (letn ((iter (lambda (hier)
                 (cond
                   ((null? hier) "")
                   ((list? hier)
                     (format "%s%s"
                       (:render (first hier) ctx)
                       (iter (rest hier) ctx)))))))
        (iter hier)))


(context MAIN)


;;; Example usage

; get template text
;(set 'text (read-file "testing/template.txt"))

;; create the view context
;(context 'data "your-name" "some body")
;(context 'data "your-email" "someone@somewhere.com")
;(context 'data "my-name" "someone else")
;(context 'data "my-email" "someone@somewhereelse.com")
;(context 'data "files" (directory "." "\\.lsp" 0))
;(context 'data "now" (date-value))

;; generate the rendered result
;(set 'template (Template:render text data))

;; if we were using the request/response modules,
;; we could send the result to the client:
;(new Response 'res)
;(print (res template))