#!/usr/bin/env newlisp


{

Usage example 1
====================
(new Response 'foo)
(foo:write "<p>Hello world!</p>")
(println (foo))
====================
Status: 200 OK
Content-type: text/html; charset=utf-8
Content-Length: 19

<p>Hello world!</p>


Usage example 2
====================
(new Response 'foo)
(println (foo "How now brown bureaucrat."))
====================
Status: 200 OK
Content-type: text/html; charset=utf-8
Content-Length: 25

How now brown bureaucrat.


Usage example 3
====================
(new Response 'foo)
(foo:redirect "/some/other/path/")
(println (foo))
====================
Status: 302 Found
Location: /some/other/path/


Usage example 4
====================
(new Response 'foo)
(println (foo:not-found "<p>These are not the droids you are looking for.</p>"))
====================
Status: 404 Not Found
Content-type: text/html; charset=utf-8
Content-Length: 52

<p>These are not the droids you are looking for.</p>


Usage example 5
====================
(new Response 'foo)
(println (foo:error "<p>An error occurred while processing this request.</p>"))
====================
Status: 500 Internal Error
Content-type: text/html; charset=utf-8
Content-Length: 55

<p>An error occurred while processing this request.</p>


}


(context 'Response)


(set '_response-codes
  '((200 "OK")
    (302 "Found")
    (404 "Not Found")
    (500 "Internal Error")))


(set '_response-code 200)
(set '_content-type "text/html; charset=utf-8")
(set '_headers '())
(set '_cookies '())
(set '_content "")


;; functions for setting headers, adding content, and predicates.


(define (Response:write)
  (doargs (a) (write-buffer _content (string a))))


(define (header key val)
  (set 'key (join (map title-case (parse key "-")) "-"))
  (if (member key _headers)
    (set-assoc (_headers key) (list key val))
    (push (list key val) _headers)))


(define (header? key)
  (lookup key _headers))


(define (headers)
  _headers)


(define (format-cookie key value domain path expires)
  ;; expires must be timestamp
  (set 'value (if value (string value) ""))
  (let ((cookie ""))
    (write-buffer cookie (format "%s=%s;" key value))
    (if domain (write-buffer cookie (format "; domain=.%s" domain)))
    (if path (write-buffer cookie (format "; path=%s" path)))
    (if expires (write-buffer cookie
      (format "; expires=%s" (date (int expires) 0 "%a, %d %b %Y %H:%M:%S %Z"))))
    cookie))


(define (set-cookie key value domain path expires)
  (if (cookie-set? key '? domain path)
    (delete-cookie key domain path))
  (push (list key value domain path expires) _cookies -1))


;; needs to check for set cookies in _cookies and remove
(define (delete-cookie key domain path)
  (if (cookie-set? key '? domain path)
    (pop _cookies (find (list key '? domain path '?) _cookies match))
    (set-cookie key nil domain path (date-value))))


(define (cookie-set? key domain path)
  (true? (find (list key '? domain path '?) _cookies match)))


;; utility functions for rendering of headers


(define (_write-cookies)
  (unless (null? _cookies)
    (let ((buffer ""))
      (dolist (cookie _cookies)
        (set 'cookie (apply format-cookie cookie))
        (write-buffer buffer (string "Set-Cookie: " cookie "\n")))
      buffer)
    ""))


(define (_write-content str)
  (if str (set '_content str))
  (if _content
    (let ((buffer ""))
      (write-buffer buffer (string "Content-type: " _content-type "\n"))
      (write-buffer buffer (string "Content-Length: " (length _content) "\n\n"))
      (write-buffer buffer (string _content))
      buffer)
    ""))


(set '_response (lambda (str)
  (let ((buffer ""))
    (let ((response-text (lookup _response-code _response-codes)))
      (write-buffer buffer
        (string "Status: " _response-code " "
          (lookup _response-code _response-codes) "\n"))
      (dolist (hdrs _headers)
        (write-buffer buffer (string (hdrs 0) ": " (hdrs 1) "\n")))
      (write-buffer buffer (_write-cookies))
      (write-buffer buffer (_write-content str))
      buffer))))


;; functions that render complete headers.  functions that need to override the
;; default rendering of complete headers simply need to replace the lambda
;; _response which returns a string containing the complete header.


(define (Response:Response str)
  (_response str))


(define (redirect path)
  (set '_response-code 302)
  (header "Location" path)
  (set '_response
    (expand 
      '(lambda ()
        (string "Status: 302 Found\n" (_write-cookies) "Location: " path 
                "\n\n"))
      'path))
  (_response))


;; 404 and 500 errors accept a string to output as the 404 page or 500 page.
;; These are useful as primitives for more sophisticated error-catching
;; routines that output formatted error and not-found pages.


(define (not-found str)
  (set '_response-code 404)
  (set '_content str)
  (_response))


(define (error str)
  (set '_response-code 500)
  (set '_content str)
  (_response))


(context MAIN)
