#!/usr/bin/env newlisp


;;; Token classes must implement :render, which should evaluate to
;;; a string.

(catch
  (define (Class:Class)
    (cons (context) (args))))


;;; Variables are simple resolvers that lookup a value in a hash.  In
;;; the event of a nil evaluation, a variable evaulates to an empty string.
;;; Filters (from FILTERS in filters.lsp) may be called by appending the
;;; variable with a | (pipe) and the name of the filter function.
;;; Arguments to a filter function are exposed by appending the function with
;;; a colon and a string argument, such as {{ variable|func:"argument" }}.

(new Class 'Variable)

(define (Variable:render self ctx)
  (letn ((parts (parse (self 1) "|"))
         (var (first parts))
         (filters (rest parts)))
        (if (context? ctx var)
            ;; If the variable exists in the view context, attempt to render
            ;; with filters
            (begin
              (set 'var (string (context ctx var)))
              (cond
                ((null? filters) var)
                (true
                  (dolist (f filters)
                    ;; parse the filter and its arguments
                    (let ((split-index (find ":" f)))
                      (let (fun func-arg)
                        (if split-index
                          (begin
                            (set 'func (0 split-index f))
                            (set 'func-arg (trim ((+ 1 split-index) f) {"})))
                          (begin
                            (set 'func f)
                            (set 'func-arg nil)))
                        ;; if the filter exists, update var, otherwise,
                        ;; safely error out
                        (if (filter? func)
                            (set 'var (string ((context FILTERS func)
                                               var func-arg)))
                            (throw-error (format "Invalid filter: %s"
                                                 func))))))))
              var)
            ;; If it does not exist in the view context, we can ignore the
            ;; filters (and not check for their existence.)
            "")))


;;; Tags use a more complex lookup procedure.  Tag functions are stored
;;; in the context TAGS.  It is documented in tags.lsp.  If a function
;;; is not found within the TAGS context for the requested, tag. render
;;; will throw an error (as opposed to a simple variable, which just
;;; returns an empty string for a nil value, since we don't know if the
;;; value was meant to be nil.)

(new Class 'Tag)

(define (Tag:extract self)
  "extract is a utility function to extract the name of the tag."
  (first (parse (self 1) " ")))


(define (Tag:args self)
  "forms a list of arguments passed to the tag function."
  (rest (parse (self 1) " ")))


(define (Tag:render self ctx)
  (let ((tag (:extract self)))
    (if (tag? tag)
        (letex ((func (context TAGS tag))
                (arg-list (:args self)))
               (string (func ctx (self 2) 'arg-list)))
        (throw-error (format "Tag not found: %s" tag)))))


;;; Strings render as... surprise! strings.

(new Class 'String)

(define (String:render self ctx)
  (string (self 1)))


;;; Comments don't render.  Really a waste of resources to use FOOP this way
;;; at all, but it makes for clearer and extensible code.

(new Class 'Comment)

(define (Comment:render self ctx) "")

